"""Configuration settings for the algobase."""


from collections.abc import Callable
from typing import Self, TypeVar

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from algobase.choices import (
    AlgorandApiProvider,
    AlgorandApiProviderChoice,
    AlgorandNetwork,
    AlgorandNetworkChoice,
)

T = TypeVar("T")


class Settings(BaseSettings):
    """Pydantic model for algobase settings."""

    model_config = SettingsConfigDict(env_prefix="AB_")

    algorand_network: AlgorandNetworkChoice = Field(
        description="The name of the Algorand network.",
        default=AlgorandNetwork.LOCALNET,
    )
    algorand_provider: AlgorandApiProviderChoice = Field(
        description="The Algorand API provider.", default=AlgorandApiProvider.LOCALHOST
    )
    algod_token: str = Field(description="The Algod API token.", default="a" * 64)

    nft_storage_api_key: str | None = Field(
        description="API key for nft.storage.", default=None
    )

    def __or__(self, f: Callable[[Self], T]) -> T:
        """Operator overloading to pipe settings into a function or other callable.

        Args:
            f (Callable[[Self], T]): The function that takes `settings` as an argument.

        Returns:
            T: The type returned by the function.
        """
        return f(self)



"""Functions for type casting."""

from collections.abc import Callable
from typing import TypeVar

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")


def maybe_bind(x: A | None, f: Callable[[A], B | None]) -> B | None:
    """Return the result of applying the function to the value if the value is not None, otherwise return None.

    Args:
        x (a | None): The value to apply the function to.
        f (Callable[[a], b  |  None]): The function to apply to the value.

    Returns:
        _type_: _description_
    """
    return f(x) if x is not None else None


def pipe(x: A, f: Callable[[A], B], g: Callable[[B], C]) -> C:
    """Pipe a value through two functions.

    ```pipe(x, f, g)``` is equivalent to ```g(f(x))```

    Args:
        x (A): The value to pipe through the functions.
        f (Callable[[A], B]): The first function to apply to the value.
        g (Callable[[B], C]): The second function to apply to the result of the first function.

    Returns:
        C: The piped value.
    """
    return g(f(x))



"""Classes and functions to configure and create Algorand API clients."""

from dataclasses import dataclass

from algosdk.kmd import KMDClient
from algosdk.v2client.algod import AlgodClient
from algosdk.v2client.indexer import IndexerClient

from algobase.choices import AlgorandApi, AlgorandApiChoice
from algobase.functional import pipe


@dataclass(frozen=True, slots=True)
class ClientConfig:
    """Configuration for an Algorand API client."""

    url: str
    credential: str
    headers: dict[str, str] | None = None


def create_algod_client(config: ClientConfig) -> AlgodClient:
    """Create an AlgodClient instance from the given configuration.

    Args:
        config (ClientConfig): The configuration to use.

    Returns:
        AlgodClient: The AlgodClient instance.
    """
    return AlgodClient(
        algod_token=config.credential, algod_address=config.url, headers=config.headers
    )


def create_indexer_client(config: ClientConfig) -> IndexerClient:
    """Create an IndexerClient instance from the given configuration.

    Args:
        config (ClientConfig): The configuration to use.

    Returns:
        IndexerClient: The IndexerClient instance.
    """
    return IndexerClient(
        indexer_token=config.credential,
        indexer_address=config.url,
        headers=config.headers,
    )


def create_kmd_client(config: ClientConfig) -> KMDClient:
    """Create a KMDClient instance from the given configuration.

    Args:
        config (ClientConfig): The configuration to use.

    Returns:
        KMDClient: The KMDClient instance.
    """
    return KMDClient(kmd_token=config.credential, kmd_address=config.url)


def create_localnet_default_config(api: AlgorandApiChoice) -> ClientConfig:
    """Create a default configuration for the localnet.

    Args:
        api (AlgorandApiChoice): The API to configure.

    Returns:
        ClientConfig: The default configuration.
    """
    port = {AlgorandApi.ALGOD: 4001, AlgorandApi.INDEXER: 8980, AlgorandApi.KMD: 7833}[
        api
    ]
    return ClientConfig(url=f"http://localhost:{port}", credential="a" * 64)


def create_localnet_algod_client() -> AlgodClient:
    """Create an AlgodClient instance for the localnet.

    Returns:
        AlgodClient: The AlgodClient instance.
    """
    return pipe(AlgorandApi.ALGOD, create_localnet_default_config, create_algod_client)


def create_localnet_indexer_client() -> IndexerClient:
    """Create an IndexerClient instance for the localnet.

    Returns:
        IndexerClient: The IndexerClient instance.
    """
    return pipe(
        AlgorandApi.INDEXER, create_localnet_default_config, create_indexer_client
    )


def create_localnet_kmd_client() -> KMDClient:
    """Create a KMDClient instance for the localnet.

    Returns:
        KMDClient: The KMDClient instance.
    """
    return pipe(AlgorandApi.KMD, create_localnet_default_config, create_kmd_client)
